#include <thread>
#include <iostream>
#include <vector>
#include <string>
#include <algorithm>
#include <set>
#include <omp.h>
#include <cstring>
#include "assert.h"
#include <fstream>

using namespace std;

void ReadBin(const string &file_path, const int node_nums,
             const int num_dimensions,
             vector<vector<int>> &data)
{
    cout << "Reading Data: " << file_path << endl;
    ifstream ifs;
    ifs.open(file_path, ios::binary);
    assert(ifs.is_open());
    cout << "# of points: " << node_nums << endl;
    vector<int> buff(num_dimensions);
    int counter = 0;
    while (ifs.read((char *)buff.data(), num_dimensions * sizeof(int)) && counter < node_nums)
    {
        vector<int> row(num_dimensions);
        for (int d = 0; d < num_dimensions; d++)
        {
            row[d] = static_cast<int>(buff[d]);
            // cout << row[d] << " ";
        }
        // cout << endl;

        data[counter++] = move(row);
    }
    ifs.close();
    cout << "Finish Reading Data" << endl;
}
int main()
{
    int K = 100;
    int compute_num = 1000000;
    int QUERY = 1000000;
    cout << "Number of compute_nodes num: " << compute_num << endl;
    vector<vector<int>> gt(compute_num), output(compute_num);
    ReadBin("../gt.bin", compute_num, 100, gt);
    ReadBin("../output.bin", compute_num, 100, output);
    int num_threads = omp_get_max_threads();
    long long int correct[5] = {0};
    long long int total[5] = {0};

    // Read queries
    string filepath_q = "../dummy-queries10.bin";
    ifstream file_q(filepath_q, ios::binary | ios::ate);
    if (!file_q.is_open())
    {
        cerr << "Failed to open file: " << filepath_q << endl;
        return -1;
    }

    streamsize size_q = file_q.tellg();
    file_q.seekg(0, ios::beg);

    vector<char> buffer_q(size_q);
    if (!file_q.read(buffer_q.data(), size_q))
    {
        cerr << "Failed to read file: " << filepath_q << endl;
        return -1;
    }
    file_q.close();

    int queryNum = QUERY;
    memcpy(&queryNum, buffer_q.data(), sizeof(int));
    cout << "queryNum=" << queryNum << endl;

    int dim_q = (size_q / 4 - 1) / queryNum;
    int real_dim_q = dim_q - 4;

    vector<float> base_q((float *)(buffer_q.data() + sizeof(int)), (float *)(buffer_q.data() + size_q));
    vector<char>().swap(buffer_q);
    cout << "query_shape=[" << queryNum << ", " << real_dim_q << "]" << endl;

    // Splitting attributes and base_q vector
    vector<int> query_type(queryNum);
#pragma omp parallel for
    for (int i = 0; i < queryNum; ++i)
    {
        query_type[i] = base_q[i * 104];
    }
    vector<float>().swap(base_q);

    // print output
    // for (int i = 0; i < compute_num; i++)
    // {
    //     cout << "gt: ";
    //     for (int j = 0; j < gt[i].size(); j++)
    //     {
    //         cout << gt[i][j] << " ";
    //     }
    //     cout << endl;
    //     cout << "output: ";
    //     for (int j = 0; j < output[i].size(); j++)
    //     {
    //         cout << output[i][j] << " ";
    //     }
    //     cout << endl;
    // }

    // #pragma omp parallel for reduction(+ : correct[ : 5], total[ : 5])
    for (int i = 0; i < compute_num; i++)
    {
        // set
        set<int> gt_set(gt[i].begin(), gt[i].end());
        // cout << "gt_set size: " << gt_set.size() << endl;
        set<int> output_set(output[i].begin(), output[i].end());
        // cout << "output_set size: " << output_set.size() << endl;
        // intersection
        vector<int> intersection;
        set_intersection(gt_set.begin(), gt_set.end(), output_set.begin(), output_set.end(), back_inserter(intersection));
        correct[0] += intersection.size();
        correct[query_type[i] + 1] += intersection.size();
        total[0] += gt_set.size();
        total[query_type[i] + 1] += gt_set.size();
    }
    // total
    cout << "Total: " << total[0] << endl;
    cout << "Total for query type 1: " << total[1] << endl;
    cout << "Total for query type 2: " << total[2] << endl;
    cout << "Total for query type 3: " << total[3] << endl;
    cout << "Total for query type 4: " << total[4] << endl;

    // correct
    cout << "Correct: " << correct[0] << endl;
    cout << "Correct for query type 1: " << correct[1] << endl;
    cout << "Correct for query type 2: " << correct[2] << endl;
    cout << "Correct for query type 3: " << correct[3] << endl;
    cout << "Correct for query type 4: " << correct[4] << endl;

    cout << "Recall@100: " << double(1.0 * correct[0] / total[0]) << endl;
    cout << "Recall@100 for query type 1: " << double(1.0 * correct[1] / total[1]) << endl;
    cout << "Recall@100 for query type 2: " << double(1.0 * correct[2] / total[2]) << endl;
    cout << "Recall@100 for query type 3: " << double(1.0 * correct[3] / total[3]) << endl;
    cout << "Recall@100 for query type 4: " << double(1.0 * correct[4] / total[4]) << endl;

    return 0;
}
